/** \file MutableDfa.java
 */

package net.sourceforge.amore.MutableLdo ;

import net.sourceforge.amore.Ldo.Dfa ;

/** intermediary data structure for editing.
 *  uses decorator pattern (cf. Gamma et al.) to delegate function calls to inner dfa.
 *  MutableDfa cannot (and should not) be used instead of Dfa in transformations, 
 *  tests, etc.
 */

public class MutableDfa {
    public final static int UNDEF = -1;
    
    Dfa d;
    public MutableDfa() {
	// default: Dfa with 1 letter alphabet, 1 state
	d = new Dfa( 1, 1 );
    }
    public MutableDfa(int sno) {
	d = new Dfa( sno, 1 );
    }

    public MutableDfa( int nletters, int nstates ) {
	d = new Dfa( nletters, nstates );
    }

    public MutableDfa(Dfa d) {
	this.d = (Dfa) d.clone();
    }
    
    /** converts this obj to Dfa. called when editing is finished
     */
    public Dfa toDfa() {
      return (Dfa) d.clone();
    }

    /** returns true if dfa is valid
     */
    public boolean isValid() {
	int qno=getNstates();
	int sno=getNletters();
	boolean val = (d.getInit() != UNDEF);
	int state=0;
	// it may not contain undef'ed transitions
	while( val & (state<qno)) { // for all states
	    int letter=1;
	    while( val & (letter<=sno) ) { // for all letters
		if( getcon( state, letter ) == UNDEF)
		    { val = false; } // ->break
		letter++;
	    }
	    state++;
	}
	return val;
    }

    // decorator methods

    // Dfa

    public void setFinal( int q )    { d.setFinal( q ); }
    public void setNonFinal( int q ) { d.setNonFinal( q ); }
    public void setInit( int q )     { d.setInit( q ); }
    public int  getInit()            { return d.getInit(); }
    public boolean isFinal( int q )  { return d.isFinalState(q); }
    public void connect(int F,int L,int T) {
	d.connect(F,L,T);
    }
    public int getcon( int F, int L ){ return d.getcon(F,L); }
    public int getNletters()         { return d.getNletters();}
    public int getNstates()          { return d.getNstates();}  

    /** copies info in this obj to (bigger) obj dest.
     *  used in addState, addLetter.
     */
  public void copyThisTo(Dfa dest) {
    int s = d.getNletters();
    int q = d.getNstates();
    System.out.println("mutableDfa.copythisto, d has "+s+" letters, "+q+" states, dest has "+dest.getNletters()+"letters, "+dest.getNstates()+" states.");
    
    /* copy initial state */
    dest.setInit( d.getInit() );
    /* copy transition table */
    for(int state = 0; state<q; state++) {      /* f. all states */
      if( d.isFinalState( state ) )
	dest.setFinal( state );
      for(int letter = 1; letter<= s; letter++) {
	dest.connect( state, letter, d.getcon( state, letter )) ;
	  }
    }
    /* copy minimal flag */
    dest.setMinimal( d.isMinimal() );
  }
    
    /** adds a new letter at the end.
     */
    public void addLetter() {
	int s = d.getNletters();
	int q = d.getNstates();
	System.out.println("mutableDfa.addLetter, d has "+s+" letters");
	Dfa dest = new Dfa( s+1, q );
	copyThisTo( dest );
	for(int i=0; i<q; i++)
	    // mark new transitions as undefined
	    dest.connect(i, s+1, UNDEF);
	
	this.d = dest;
    }
    
    /** removes last letter
     */
    public void removeLetter() {
	Dfa dest;
	int s = d.getNletters();
	System.out.println("mutableDfa.removeLetter, d has "+s+" letters");
	if (s > 1) {
	    int q = d.getNstates();
	    dest = new Dfa( s - 1, q );
	    dest.setInit( d.getInit() );
	    for(int i = 0; i<q; i++) {
		if( d.isFinalState( i ) )
		    dest.setFinal( i );
		for(int j = 1; j<= s-1; j++) /* letter */
		    dest.connect( i, j, d.getcon(i,j));
	    } /* for..*/	  
	    this.d = dest;
	} /* if.. */
	
    }
    
    /** adds a new state at the end.
     */
    public void addState() {
	int s = d.getNletters();
	int q = d.getNstates();
	Dfa dest = new Dfa( s, q + 1);
	copyThisTo( dest );
	for(int i=1; i<=s; i++) { // all letters starting from 'a'
	  // mark all transitions as undefined
	  // new state has index q 
	  dest.connect(q, i, UNDEF);
	}
	this.d = dest;
    }
    
    /** removes last state ???
     */
    public void removeState() {
	Dfa dest;
	int qno = d.getNstates();
	if (qno > 1) {
	    int s = d.getNletters();
	    dest = new Dfa( s, qno - 1 );
	    //  if the init state is deleted, mark as UNDEF
	    int init = d.getInit();
	    if( init != qno )
		dest.setInit( init );
	    else
		dest.setInit( UNDEF );
	    System.out.println("hallo");
	    for(int i = 0; i< qno - 1; i++) {
		if( d.isFinalState( i ) )
		    dest.setFinal( i );
		
		for(int j = 1; j<=s; j++) {/* letter */
		    int target = d.getcon(i,j);

		    if (target < qno - 1)
			dest.connect( i, j, target );
		    else
			// mark as UNDEF
			dest.connect( i, j, UNDEF );
			
		}
	    } /* for..*/
	    this.d = dest;
	} /* if..*/
    }
    
  public String toString() {
      return("<-- MutableDfa containing "+d.toString()+"-->");
  }
    
}
